// $Id: zload.c 35 2006-06-29 23:39:02Z Don Kinzer $

/*
 ** Module: zload.c
 *
 * Copyright 2005 Elba Corp.
 * All rights reserved.
 *
 * Non-exclusive permission to use, copy, modify, and distribute this
 * software, without fee, for the limited purpose of loading code into
 * the ZX-series processors is hereby granted, provided:
 *   1) that the above copyright notice is retained in the source code and,
 *   2) that it is understood and agreed that no warranty of any kind,
 *      express or implied is provided.
 * Use of this code in any manner shall be prima facie evidence of
 * acceptance of these terms.
 *
 *
 * This module implements a downloader/monitor that works with the ZX-
 * series microcontrollers.  The various options allow you to send an Intel
 * hex file to the microcontroller, restart the microcontroller and then
 * monitor the output in terminal mode.  You can also perform a firmware
 * upgrade on the ZX processor.
 *
 * The code was written specifically for Microsoft Windows platforms but it
 * may be able to be modified to run on other platforms as well.
 *
 */

/** include files **/
#ifdef WIN32
#include <windows.h>
#include <conio.h>
#endif
#include <stdio.h>
#include <io.h>
#include <ctype.h>
#include <limits.h>
#include "zxcomm.h"

/** local definitions **/

#define ACK_CHAR				'A'			// positive acknowledgement
#define NAK_CHAR				'N'			// negative acknowledgement

typedef struct
{
	int quiet;
	int dotCnt;
	long retryCnt;
	long mask;			// for progress updates, the mask controlling update frequency
	FILE *ofp;
} DL_data;

/** private data **/
static unsigned verMajor = 1;
static unsigned verMinor = 1;
static unsigned verVariant = 5;

/** internal functions **/
static void commDeInit(HANDLE hand);
static void displayHelp(void);
static void doAtnCheck(unsigned port);
static void openComm(unsigned port, HANDLE *handp);
static int getVal(const char **bufp, unsigned long *valp, int radix);
static void resumeNormalSpeed(HANDLE cHand);
static void runTest(int mode, unsigned port);

/** callback functions **/
void CALLBACK mesgFunc(const char *mesg, long mesgData);
WORD CALLBACK progFunc(long recNum, long recCnt, long retryCnt, long data);

/** public functions **/

int
main(int argc, char **argv)
{
	int i = 0;
	HANDLE cHand = INVALID_HANDLE_VALUE;
	unsigned port = 1;
	int testMode = 0;
	int quiet = FALSE;
	int verify = FALSE;
	int identify = FALSE;
	int termMode = FALSE;
	COMSTAT cs;
	DWORD cnt;
	char termChar = 0x04;
	int checkTerm = FALSE;
	int downloadCount = 0;
	BYTE doUpdate = 0;
	char cfgBuf[50];
	int cfgBufLen = 0;
	unsigned recCnt = 0;

	// set the default port
	const char *p;
	if ((p = getenv("ZLOAD_PORT")) == NULL)
		p = "1";
	port = atoi(p);

	argv++;
	while (--argc > 0)
	{
		char c;
		const char *p, *argp;

		argp = *argv++;
		if ((c = *argp) == '-')
		{
			c = argp[1];
			p = argp + 2;
			switch(c)
			{
			case 'a':
				doAtnCheck(port);
				exit(0);
				break;

			case 'c':
				if (((port = atoi(p)) <= 0) || (port > 99))
				{
					fprintf(stderr, "Invalid comm channel: \"%s\"\n", argp);
					exit(1);
				}
				break;

			case 's':
				checkTerm = TRUE;
				if ((c = *p) == '\0')
					termChar = 0x04;
				else if ((c == '0') && (((c = p[1]) == 'X') || (c == 'x')))
				{
					termChar = 0;
					for (p += 2; (c = *p++) != '\0'; )
					{
						char val;

						if (isxdigit(c))
							val = hexVal(c);
						else
						{
							fprintf(stderr, "Invalid character for hexadecimal value - \"%s\"\n", argp);
							exit(1);
						}
						termChar = (termChar * 16) + val;
					}
				}
				else
				{
					termChar = 0;
					while ((c = *p++) != '\0')
					{
						if (isdigit(c))
							termChar = (termChar * 10) + c - '0';
						else
						{
							fprintf(stderr, "Invalid character for decimal value - \"%s\"\n", argp);
							exit(1);
						}
					}
				}
				break;

			case 'i':
				identify = TRUE;
				break;

			case 'm':
				termMode = TRUE;
				break;

			case 'v':
				verify = TRUE;
				break;

			case 'q':
				quiet = TRUE;
				break;

			case 'e':		// perform emergency update
			case 'u':		// perform normal update
			case 'U':		// perform update without blinking lights
				doUpdate = c;
				break;

			case 'z':		// device configuration, z<code>:<value>
				c = *p++;
				if (isdigit(c) && (*p++ == ':') && (*p != '\0'))
				{
					int radix = 10;
					unsigned long val;
					char cfg = c;

					// read the configuration value
					if ((*p == '0') && (((c = p[1]) == 'x') || (c == 'X')))
						p += 2, radix = 16;
					if (getVal(&p, &val, radix) != 0)
					{
						fprintf(stderr, "Invalid character in %sdecimal value - \"%s\"\n",
								(radix == 16) ? "hexa" : "", argp);
						exit(1);
					}

					if ((cfg == '0') && ((cfgBufLen + 3) < sizeof(cfgBuf)))
					{
						// add configuration data to the buffer for later use
						cfgBuf[cfgBufLen++] = cfg - '0';
						cfgBuf[cfgBufLen++] = (char)val;
						cfgBuf[cfgBufLen++] = (char)(val >> 8);
						break;
					}
				}
				fprintf(stderr, "Invalid device configuration - \"%s\"\n", argp);
				exit(1);
				break;

			case 't':
				testMode = atoi(argp + 2) + 1;
				break;

			case '?':
			case 'h':
			case 'H':
				displayHelp();
				exit(0);
				break;

			case '#':
				{
					int radix = 10;
					unsigned long val;

					if (((c = *p) == 'x') || (c == 'X'))
						p++, radix = 16;
					if (getVal(&p, &val, radix) != 0)
					{
						fprintf(stderr, "Invalid character in %sdecimal value - \"%s\"\n",
								(radix == 16) ? "hexa" : "", argp);
						exit(1);
					}
					setDiagCode(val);
				}
				break;

			default:
				fprintf(stderr, "Unrecognized option - \"%s\".\n", argp);
				exit(1);
				break;
			}
		}
		else if ((c == '/') && (((c = argp[1]) == 'H') || (c == 'h') || (c == '?')))
		{
			displayHelp();
			exit(0);
		}
		else if (doUpdate)
		{
			DL_data dld;
			int stat;
			int flags;
			if (doUpdate == 'e')
				flags = ZX_UPDATE_EMERGENCY;
			else if (doUpdate == 'U')
				flags = ZX_UPDATE_NO_BLINK;
			else
				flags = ZX_UPDATE_NORMAL;

			// prepare the progress update parameters
			dld.quiet = quiet;
			dld.dotCnt = 0;
			dld. retryCnt = 0;
			dld.mask = 0x1;
			dld.ofp = stderr;

			// effect the download
			stat = ZXUpdateEx(argp, port, flags, progFunc, (long)&dld, mesgFunc, (long)&dld);
			exit((stat == ZX_SUCCESS) ? 0 : 1);
		}
		else
		{
			DL_data dld;
			int flags = ZX_DOWNLOAD_WRITE | ZX_NO_RESET | ZX_COMM_HANDLE;
			if (verify)
				flags |= ZX_DOWNLOAD_VERIFY;
			openComm(port, &cHand);

			// prepare the progress update parameters
			dld.quiet = quiet;
			dld.dotCnt = 0;
			dld. retryCnt = 0;
			dld.mask = 0x1f;
			dld.ofp = stderr;

			// effect the download
			if (ZXDownloadEx(argp, (long)cHand, flags, progFunc, (long)&dld, mesgFunc, (long)&dld) != ZX_SUCCESS)
				exit(1);
			downloadCount++;
		}
	}

	if (downloadCount)
	{
		// allow time for the device to finish sending any response
		msDelay(500);

		// empty the incoming queue
		clearQueue(cHand);

		if (termMode)
			resumeNormalSpeed(cHand);

		// reset the device to cause it to begin execution
		doReset(cHand);
	}
	else if (cfgBufLen)
	{
		DL_data dld;
		int flags = ZX_COMM_HANDLE | ZX_NO_RESET;
		openComm(port, &cHand);

		// prepare the progress update parameters
		dld.quiet = FALSE;
		dld.dotCnt = 0;
		dld. retryCnt = 0;
		dld.mask = 0;
		dld.ofp = stderr;

		// perform the configuration
		ZXConfigure(cfgBuf, cfgBufLen, (long)cHand, flags, mesgFunc, (long)&dld);

		if (termMode)
			resumeNormalSpeed(cHand);

		// reset the device to cause it to begin execution
		doReset(cHand);
	}
	else if (identify)
	{
		DL_data dld;
		int flags = ZX_DOWNLOAD_IDENTIFY | ZX_COMM_HANDLE | ZX_NO_RESET;
		openComm(port, &cHand);

		// prepare the progress update parameters
		dld.quiet = FALSE;
		dld.dotCnt = 0;
		dld. retryCnt = 0;
		dld.mask = 0;
		dld.ofp = stderr;

		// effect the download
		if (ZXIdentify((long)cHand, flags, mesgFunc, (long)&dld) != ZX_SUCCESS)
			exit(1);

		if (termMode)
			resumeNormalSpeed(cHand);

		// reset the device to cause it to begin execution
		doReset(cHand);
	}
	else if (!termMode && !testMode)
	{
		fprintf(stderr, "No file specified to download.  Use -h for help.\n");
		exit(1);
	}

	if (testMode)
		runTest(testMode, port);
	else if (termMode)
	{
		// make sure the comm port is open
		if (cHand == INVALID_HANDLE_VALUE)
			openComm(port, &cHand);

#ifdef WIN32x
		// set console input to "raw" mode
		{
			HANDLE h = GetStdHandle(STD_INPUT_HANDLE);
			DWORD mode;
			if ((h != INVALID_HANDLE_VALUE) && GetConsoleMode(h, &mode))
			{
				mode &= ~ENABLE_LINE_INPUT;
				SetConsoleMode(h, mode);
			}
		}
#endif

		// echo incoming characters to the console
		fflush(stdin);
		while (1)
		{
			char c;

			ClearCommError(cHand, &cnt, &cs);
			if (cs.cbInQue == 0)
			{
				// check for a console key being available
				if (_kbhit())
				{
					DWORD cnt;
					if ((c = _getch()) == '\r')
						c = '\n';
					WriteFile(cHand, &c, 1, &cnt, NULL);
				}
				Yield();
			}
			else if (ReadFile(cHand, &c, 1, &cnt, NULL) && (cnt > 0))
			{
				if (checkTerm && (c == termChar))
					break;
				if (c != '\r')
				{
					fputc(c, stdout);
					fflush(stdout);
				}
			}
		}
	}

	if (cHand != INVALID_HANDLE_VALUE)
		commDeInit(cHand);
	exit(0);
	return(0);
}

/** callback functions **/

/*
 * This callback outputs the supplied message, adding a CRLF.
 *
 */
void CALLBACK
mesgFunc(const char *mesg, long mesgData)
{
	DL_data *dlp = (DL_data *)mesgData;
	if ((mesg != NULL) && (dlp != NULL) && (dlp->ofp != NULL))
	{
		if (dlp->dotCnt)
			fputs("\n", dlp->ofp);
		dlp->dotCnt = 0;
		fputs(mesg, dlp->ofp);
		fputs("\n", dlp->ofp);
		fflush(dlp->ofp);
	}
}

/*
 * This callback outputs a dot to indicate progress.  The 'mask'
 * value controls the ratio of calls to dots, essentially being
 * a modulus operation.  A CRLF is output periodically to avoid
 * having the line of dots grow too long.
 *
 */
WORD CALLBACK
progFunc(long recNum, long recCnt, long retryCnt, long progData)
{
	DL_data *dlp = (DL_data *)progData;
	if ((dlp != NULL) && (dlp->ofp != NULL) && !dlp->quiet)
	{
		if ((recNum & dlp->mask) == 0)
		{
			fputc((dlp->retryCnt == retryCnt) ? '.' : '!', dlp->ofp);
			dlp->retryCnt = retryCnt;
			if (++dlp->dotCnt > 72)
			{
				if (recCnt != 0)
					fprintf(dlp->ofp, " %u%%", (recNum * 100 + 50) / recCnt);
				fputc('\n', dlp->ofp);
				dlp->dotCnt = 0;
			}
		}
		else if ((recNum == recCnt) && dlp->dotCnt)
		{
			fputs(" 100%%\n", dlp->ofp);
			dlp->dotCnt = 0;
		}
		fflush(dlp->ofp);
	}
	return(1);
}

/** private functions **/

/*
 ** commDeInit
 *
 * Clean up after a comm session.
 *
 */
static void
commDeInit(HANDLE hand)
{
	if (hand != INVALID_HANDLE_VALUE)
		CloseHandle(hand);
}

/*
 ** displayHelp
 *
 * Display invocation help on stdout.
 *
 */
static void
displayHelp(void)
{
	char verStr[20];
	int len;

	// compose the version number
	len = sprintf(verStr, "%u.%u", verMajor, verMinor);
	if (verVariant)
		sprintf(verStr + len, ".%u", verVariant);

	fprintf(stdout, "Invocation:          (V%s)\n", verStr);
	fprintf(stdout, "  zload [<options>]  [<file>]\n");
	fprintf(stdout, "    where <options>  are:\n");
	fprintf(stdout, "    -h               display this information\n");
	fprintf(stdout, "    -a               enter ATN test mode\n");
	fprintf(stdout, "    -c<port>         specify COM port by a decimal value\n");
	fprintf(stdout, "    -i               retrieve an identification string from the device\n");
	fprintf(stdout, "    -m               after downloading continue running in monitor mode\n");
	fprintf(stdout, "    -s[<charCode>]   exit monitor mode when a character code is received\n");
	fprintf(stdout, "    -u               perform a firmware update using the specified file\n");
	fprintf(stdout, "    -U               same as above but without visual feedback\n");
	fprintf(stdout, "    -v               after downloading perform a verification\n");
	fprintf(stdout, "    -q               don't provide progress feedback\n");
	fprintf(stdout, "    -z<code>:<value> set device configuration\n");
	fprintf(stdout, "Example: zload -c2 -s0x23 -m test.zxb\n");
}

/*
 ** doAtnCheck
 *
 * Attempt to gain control of the client system by toggling the DTR line.
 *
 * Return zero if successful, non-zero otherwise.
 *
 */
static void
doAtnCheck(unsigned port)
{
	int i = 0;
	HANDLE cHand;
	char buf[50];

	// try to open the comm port
	openComm(port, &cHand);

	fputs("\n", stdout);
	fprintf(stdout, "                ATN Test Mode\n");
	fprintf(stdout, "The DTR line will be toggled each time you press Enter.\n");
	fprintf(stdout, "Press Ctrl-Break to exit when finished testing.\n");
	fputs("\n", stdout);
	while (1)
	{
		if (i & 1)
		{
			EscapeCommFunction(cHand, SETDTR);
			fprintf(stdout, "The voltage on pin 3 should now be +3 to +12 volts.\n");
		}
		else
		{
			EscapeCommFunction(cHand, CLRDTR);
			fprintf(stdout, "The voltage on pin 3 should now be -3 to -12 volts.\n");
		}
		fgets(buf, sizeof(buf), stdin);
		fputs("\n", stdout);
		i++;
	}
	commDeInit(cHand);
}

//
// Read a numeric value from a buffer using the specified radix.  Return
// the value and a pointer to the first invalid character indirectly.
// Return true if one or more valid digits were read before encountering
// a null in the buffer.  Also, false is returned if an invalid character
// or an invalid digit for the radix is encountered.
//
// Radices between 2 and 10 (inclusively) are supported as well as 16.
//
static int
getVal(const char **bufp, unsigned long *valp, int radix)
{
	int digitCnt = 0;
	unsigned long val = 0;
	const char *p;
	int stat = -1;

	// validate the requested radix
	if ((radix < 2) || ((radix > 10) && (radix != 16)))
		return(stat);
	if ((bufp == NULL) || ((p = *bufp) == NULL))
		return(stat);
	for (val = 0; ; p++)
	{
		char c;
		if ((c = *p) == '\0')
		{
			if (digitCnt > 0)
				stat = 0;
			break;
		}
		if (isdigit(c))
			c -= '0';
		else if ((radix == 16) && isxdigit(c))
		{
			if (isupper(c))
				c -= ('A' - 10);
			else
				c -= ('a' - 10);
		}
		else
			// invalid character
			break;

		if (c >= radix)
			// invalid digit for the radix
			break;

		// add the digit to the value accumulator
		val = (val * radix) + c;
		digitCnt++;
	}
	*bufp = p;
	*valp = val;
	return(stat);
}

/*
 ** openComm
 *
 * Open a serial port.
 *
 */
static void
openComm(unsigned port, HANDLE *handp)
{
	if (commInit(port, handp) != ZX_SUCCESS)
	{
		fprintf(stderr, "Can't open COM%u.", port);
		exit(1);
	}
}

/*
 ** resumeNormalSpeed
 *
 * Switch the comm port back to normal speed.
 *
 */
static void
resumeNormalSpeed(HANDLE cHand)
{
	if ((DL_SPEED != STD_SPEED) && (setSpeed(cHand, STD_SPEED) != 0))
	{
		DWORD err = GetLastError();
		fprintf(stderr, "Attempt to change back to normal speed failed, err=%lu.\n", err);
		exit(1);
	}
}

/*
 ** runTest
 *
 * This function provides various diagnostic tests that may be useful
 * during the development process - not intended for end-user use.
 *
 */
static void
runTest(int mode, unsigned port)
{
	HANDLE cHand = INVALID_HANDLE_VALUE;

	switch(mode)
	{
	case 1:		// toggle ATN continuously
		commInit(port, &cHand);
		if (cHand == INVALID_HANDLE_VALUE)
			exit(1);
		while (1)
		{
			msDelay(1);
			atn(cHand, 100);
		}
		break;

	case 2:		// send sequential characters, check for ack for each 16-byte block
		commInit(port, &cHand);

		EscapeCommFunction(cHand, CLRDTR);
		if (DL_SPEED != STD_SPEED)
			setSpeed(cHand, DL_SPEED);
		if (doAtn(cHand, TRUE) == 0)
		{
			DWORD cnt;
			BYTE c, bcnt;
			COMSTAT cs;
			unsigned acnt, ncnt;
			WriteFile(cHand, "@\r", 2, &cnt, NULL);
			msDelay(250);
			clearQueue(cHand);
			bcnt = 0;
			acnt = ncnt = 0;
			for (c = 0; ; c++)
			{
				WriteFile(cHand, &c, 1, &cnt, NULL);
				if (++bcnt == 16)
				{
					char resp;
					while (1)
					{
						ClearCommError(cHand, &cnt, &cs);
						if (cs.cbInQue != 0)
							break;
						Yield();
					}
					if (ReadFile(cHand, &resp, 1, &cnt, NULL) && (cnt > 0))
					{
						char *str;
						if (resp == ACK_CHAR)
						{
							acnt++;
							str = "ACK";
						}
						else if (resp == NAK_CHAR)
						{
							ncnt++;
							str = "NAK";
						}
						else
						{
							fprintf(stderr, "unexpected device response - 0x%02x\n", (int)resp & 0xff);
							str = NULL;
						}
						if (str != NULL)
							fprintf(stderr, "%s 0x%02x, #ack=%u, #nak=%u\n", str, (int)c & 0xff, acnt, ncnt);
					}
					bcnt = 0;
				}
				Yield();
				usDelay(1000);
			}
		}
		break;
	}
	if (cHand != INVALID_HANDLE_VALUE)
		commDeInit(cHand);
}

