// $Id: zxcomm.c 35 2006-06-29 23:39:02Z Don Kinzer $

/*
 ** Module: zxcomm.c
 *
 * Copyright 2005 Elba Corp.
 * All rights reserved.
 *
 * Non-exclusive permission to use, copy, modify, and distribute this
 * software, without fee, for the limited purpose of loading code into
 * the ZX-series processors is hereby granted, provided:
 *   1) that the above copyright notice is retained in the source code and,
 *   2) that it is understood and agreed that no warranty of any kind,
 *      express or implied, is provided.
 * Use of this code in any manner shall be prima facie evidence of
 * acceptance of these terms.
 *
 *
 * This module implements device communication functions for interacting
 * with ZX- series microcontrollers.  The various functions allow you to send
 * an Intel hex file to the microcontroller, restart the microcontroller, or
 * perform a firmware upgrade on the ZX processor.
 *
 * The code was written specifically for Microsoft Windows platforms but it
 * may be able to be modified to run on other platforms as well.
 *
 */

/** include files **/
#ifdef WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#include "zxcomm.h"

/** local definitions **/

#define HEX_READ_PROMPT			0
#define HEX_READ_START			1
#define HEX_READ_BYTECNT		2
#define HEX_READ_ADDRESS		3
#define HEX_READ_RECTYPE		4
#define HEX_READ_RECORD			5
#define HEX_READ_CHKSUM			6
#define HEX_READ_DONE			100

#define HEX_WRITE_PROMPT		0
#define HEX_WRITE_RECORDS		1
#define HEX_WRITE_DONE			100

#define REC_TYPE_DATA			0			// Program Memory data
#define REC_TYPE_END			1			// end record
#define REC_TYPE_RAM			0x55		// RAM
#define REC_TYPE_EEPROM			0x56		// EEPROM data
#define REC_TYPE_REGISTER		0x57		// register data
#define REC_TYPE_FIRMWARE		0x58		// firmware version

#define REC_TYPE_INVALID		0xff

#define PROMPT					'>'

// acknowledgement characters
#define NO_ACK					'\0'
#define ESC_CHAR				0x1b		// abort command
#define ACK_CHAR				'A'			// positive acknowledgement
#define NAK_CHAR				'N'			// negative acknowledgement
#define VER_CHAR				'V'			// verification error

#define RST_CHAR				'R'			// received Reset record
#define EOF_CHAR				'E'			// received EOF record
#define SIG_CHAR				'S'			// signature mismatch
#define CRC_CHAR				'C'			// CRC mismatch
#define TIMEOUT_CHAR			'T'			// timed out awaiting character

#define MAX_REC_LEN				43			// 16 bytes plus hex format overhead

#define ATN_RETRY				5			// number of times to try to get an ATN response
#define ATN_COUNT				6			// number of times to toggle DTR for ATN
#define ATN_TIMEOUT				250			// number of milliseconds to await response

#define RST_TIME				100			// reset pulse width, in microseconds
#define RST_COUNT				2			// number of times to toggle DTR for ordinary reset

// diagnostic codes for debugging
#define DIAG_NO_TIME_LIMIT		0x0001
#define DIAG_VERBOSE			0x0002

#define CLOSE_PORT				0x0001

// this structure provides data that is used internally to realize the operations
typedef struct
{
	HANDLE hand;		// the serial port to use
	unsigned flags;		// operation flags
	const char *file;	// if not null, the file to send
	FILE *ifp;			// if not null, the handle to the above file
	char *buf;			// a buffer containing download data
	unsigned bufLen;	// number of bytes in the buffer
	long recCnt;		// the number of records in the file or buffer
	ProgFunc progFunc;	// if not null, a function to call to update the progress
	long progData;		// additional parameter to send to progFunc
	MesgFunc mesgFunc;	// if not null, a function to call to display error messages
	long mesgData;		// additional parameter to send to mesgFunc
} ZXCommData;

typedef struct
{
	HWND hWnd;			// the window to which to send a message
	UINT mesg;			// the message to send
	LPARAM lParam;		// the lParam to send with the message
	long flags;
} CBData;

/** private data **/
static int diagCode = 0;		// for debugging purposes

/** public data **/

/** internal functions **/
static int commOpen(unsigned port, ZXCommData *zdp);
static int configureDevice(HANDLE hand, const char *cfgData, int cfgLen);
static unsigned short crc(unsigned char ch, unsigned short prevCRC);
static int generateHexRecord(char *recBuf, const char *dataBuf, unsigned cnt, unsigned addr, char recType);
static BYTE getRecordByte(const BYTE *buf, int idx);
static int getResponse(ZXCommData *zdp, long recCnt);
static int identifyDevice(ZXCommData *zdp);
static int prepareForDownload(ZXCommData *zdp, int needAtnResponse);
static void resetDevice(HANDLE hand);
static int sendHexRecords(ZXCommData *zdp, unsigned flags);
static char *skipWhite(const char *p);
static int trimWhite(char *str);
static int updateFirmware(ZXCommData *zdp, unsigned flags, int *retryp);
static void updateWindows(unsigned maxMsgs);
static BYTE validateHexRecord(const BYTE *buf, int *lenp);
static int verifyIntelHex(ZXCommData *zdp);
static int verifyUpdate(ZXCommData *zdp);

/** callback functions **/
WORD CALLBACK ZXHWndProgressCB(long recNum, long recCnt, long retryCnt, long data);

/** exported functions **/

/*--------------------------------------------------------------------------*/

/*
 ** ZXDownloadEx
 *
 * This exported entry point provides a means to download the contents of an
 * Intel Hex file to the ZX program memory.  After each record is sent, the
 * 'progFunc' callback is called (if not null), passing it the current record
 * number and the total number of records.  This allows a progress indicator
 * to be periodically updated.  If the 'mesgFunc' callback is not null, it
 * will be called with message strings, typically to indicate an error condition
 * but also to report operation completion.
 *
 * If the process completes successfully, the function will return ZX_SUCCESS.
 * Otherwise, a value will be returned to indicate the error condition (see
 * zxcomm.h for a list of values and their meanings).
 *
 */
int CALLBACK
ZXDownloadEx(const char *file, long port, long flags, ProgFunc progFunc, long progData, MesgFunc mesgFunc, long mesgData)
{
	int stat = ZX_SUCCESS;
	char msg[200];
	ZXCommData zxd;

	memset(&zxd, 0, sizeof(zxd));
	zxd.hand = INVALID_HANDLE_VALUE;
	zxd.file = file;
	zxd.progFunc = progFunc;
	zxd.progData = progData;
	zxd.mesgFunc = mesgFunc;
	zxd.mesgData = mesgData;

	// attempt to open the specified file
	if ((file == NULL) || (*file == '\0'))
		return(ZX_PARAMETER_ERROR);
	if ((zxd.ifp = fopen(file, "r")) == NULL)
	{
		if (zxd.mesgFunc)
		{
			sprintf(msg, "Can't open file \"%s\" for reading.", file);
			(*zxd.mesgFunc)(msg, zxd.mesgData);
		}
		return(ZX_ERROR_FILE_OPEN);
	}

	// confirm that the file contains a valid download image
	if (!verifyIntelHex(&zxd))
	{
		if (zxd.mesgFunc)
		{
			sprintf(msg, "File \"%s\" does not contain a proper download image.", file);
			(*zxd.mesgFunc)(msg, zxd.mesgData);
		}
		stat = ZX_INVALID_DOWNLOAD_FILE;
		goto done;
	}

	// attempt to open the specified serial port
	if (flags & ZX_COMM_HANDLE)
		zxd.hand = (HANDLE)port;
	else if ((stat = commOpen(port, &zxd)) == ZX_SUCCESS)
		zxd.flags |= CLOSE_PORT;
	else
		goto done;

	// get the device into the load mode
	if ((stat = prepareForDownload(&zxd, TRUE)) != ZX_SUCCESS)
		goto done;

	// if requested, get the device identification string
	if ((flags & ZX_DOWNLOAD_IDENTIFY) && mesgFunc)
		identifyDevice(&zxd);

	// download the image
	if (flags & ZX_DOWNLOAD_WRITE)
		stat = sendHexRecords(&zxd, ZX_DOWNLOAD_WRITE);

	// allow time for the device to finish sending any response
	msDelay(250);

	// empty the incoming queue
	clearQueue(zxd.hand);

	// perform verification
	if (flags & ZX_DOWNLOAD_VERIFY)
		stat = sendHexRecords(&zxd, ZX_DOWNLOAD_VERIFY);

	// reset the device to cause it to begin executing
	if ((flags & ZX_NO_RESET) == 0)
		doReset(zxd.hand);

done:
	// clean up
	if (zxd.ifp != NULL)
		fclose(zxd.ifp);
	if ((zxd.flags & CLOSE_PORT) && (zxd.hand != INVALID_HANDLE_VALUE))
		CloseHandle(zxd.hand);
	return(stat);
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXDownload
 *
 * This exported entry point provides a means to download the contents of an
 * Intel Hex file to the ZX program memory.  It is identical to ZXDownloadEx
 * except that no means is provided to pass message strings back to the caller.
 *
 */
int CALLBACK
ZXDownload(const char *file, long port, long flags, ProgFunc progFunc, long progData)
{
	return(ZXDownloadEx(file, port, flags, progFunc, progData, NULL, 0));
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXUpdateEx
 *
 * This exported entry point provides a means to download the contents of a
 * specially formatted firmware update to the ZX.  After each record is sent, the
 * 'progFunc' callback is called (if not null), passing it the current record
 * number and the total number of records.  This allows a progress indicator
 * to be periodically updated.  If the 'mesgFunc' callback is not null, it
 * will be called with message strings, typically to indicate an error condition
 * but also to report operation completion.
 *
 * If the process completes successfully, the function will return ZX_SUCCESS.
 * Otherwise, a value will be returned to indicate the error condition (see
 * zxcomm.h for a list of values and their meanings).
 *
 */
int CALLBACK
ZXUpdateEx(const char *file, long port, long flags, ProgFunc progFunc, long progData, MesgFunc mesgFunc, long mesgData)
{
	int stat = ZX_SUCCESS;
	int retries;
	char msg[200];
	ZXCommData zxd;

	memset(&zxd, 0, sizeof(zxd));
	zxd.hand = INVALID_HANDLE_VALUE;
	zxd.file = file;
	zxd.progFunc = progFunc;
	zxd.progData = progData;
	zxd.mesgFunc = mesgFunc;
	zxd.mesgData = mesgData;

	// attempt to open the specified file
	if ((file == NULL) || (*file == '\0'))
		return(ZX_PARAMETER_ERROR);
	if ((zxd.ifp = fopen(file, "r")) == NULL)
	{
		if (zxd.mesgFunc)
		{
			sprintf(msg, "Can't open file \"%s\" for reading.", file);
			(*zxd.mesgFunc)(msg, zxd.mesgData);
		}
		return(ZX_ERROR_FILE_OPEN);
	}

	// confirm that the file contains a valid download image
	if (!verifyUpdate(&zxd))
	{
		if (zxd.mesgFunc)
		{
			sprintf(msg, "Update file \"%s\" failed integrity check.", file);
			(*zxd.mesgFunc)(msg, zxd.mesgData);
		}
		stat = ZX_INVALID_UPDATE_FILE;
		goto done;
	}

	// attempt to open the specified serial port
	if (flags & ZX_COMM_HANDLE)
		zxd.hand = (HANDLE)port;
	else if ((stat = commOpen(port, &zxd)) == ZX_SUCCESS)
		zxd.flags |= CLOSE_PORT;
	else
		goto done;

	// get the device into the load mode
	if ((stat = prepareForDownload(&zxd, TRUE)) != ZX_SUCCESS)
		goto done;

	// download the image
	retries = 0;
	stat = updateFirmware(&zxd, flags & ZX_MODE_MASK, &retries);
	if ((stat == ZX_SUCCESS) && zxd.mesgFunc)
	{
		char retryMsg[20];
		if (retries == 0)
			*retryMsg = '\0';
		else
			sprintf(retryMsg, " (with %u retr%s)", retries, (retries == 1) ? "y": "ies");
		sprintf(msg, "Firmware update completed successfully%s.", retryMsg);
		(*zxd.mesgFunc)(msg, zxd.mesgData);
	}

	// reset the device to cause it to begin executing
	if ((flags & ZX_NO_RESET) == 0)
		doReset(zxd.hand);

done:
	// clean up
	if (zxd.ifp != NULL)
		fclose(zxd.ifp);
	if ((zxd.flags & CLOSE_PORT) && (zxd.hand != INVALID_HANDLE_VALUE))
		CloseHandle(zxd.hand);
	return(stat);
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXUpdate
 *
 * This exported entry point provides a means to download the contents of an
 * Intel Hex file to the ZX program memory.  It is identical to ZXUpdateEx
 * except that no means is provided to pass message strings back to the caller.
 *
 */
int CALLBACK
ZXUpdate(const char *file, long port, long flags, ProgFunc progFunc, long progData)
{
	return(ZXUpdateEx(file, port, flags, progFunc, progData, NULL, 0));
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXConfigure
 *
 * This exported entry point provides a means to send configuration data to
 * the ZX.  If the 'mesgFunc' callback is not null, it will be called with
 * message strings, typically to indicate an error condition but also to
 * report operation completion.
 *
 * If the process completes successfully, the function will return ZX_SUCCESS.
 * Otherwise, a value will be returned to indicate the error condition (see
 * zxcomm.h for a list of values and their meanings).
 *
 */
int CALLBACK
ZXConfigure(const char *cfgData, long cfgLen, long port, long flags, MesgFunc mesgFunc, long mesgData)
{
	int stat = ZX_SUCCESS;
	unsigned bufLen;
	char msg[100];
	ZXCommData zxd;

	if ((cfgData == NULL) || (cfgLen == 0))
		return(ZX_PARAMETER_ERROR);

	memset(&zxd, 0, sizeof(zxd));
	zxd.hand = INVALID_HANDLE_VALUE;
	zxd.mesgFunc = mesgFunc;
	zxd.mesgData = mesgData;

	// attempt to open the specified serial port
	if (flags & ZX_COMM_HANDLE)
		zxd.hand = (HANDLE)port;
	else if ((stat = commOpen(port, &zxd)) == ZX_SUCCESS)
		zxd.flags |= CLOSE_PORT;
	else
		goto done;

	// compute the size of buffer required
	bufLen = (cfgLen + 15) / 16 * 16;	// number of records
	bufLen = (bufLen * 2) +	12;			// maximum size for hex record form
	bufLen += 12;						// size of the end record

	// allocate a buffer for the hex records
	if ((zxd.buf = malloc(bufLen + 20)) == NULL)
	{
		stat = ZX_ALLOC_ERROR;
		goto done;
	}

	// generate hex records for the data
	while (cfgLen && (zxd.bufLen < bufLen))
	{
		unsigned addr = 0;
		int cnt, dlen;
		char buf[20];
		char recType;

		switch(cfgData[0])
		{
		case 0:		// EEPROM configuration
			if (cfgLen < 3)
			{
				if (zxd.mesgFunc)
				{
					sprintf(msg, "missing configuration data - %u bytes\n", cfgLen);
					(*zxd.mesgFunc)(msg, zxd.mesgData);
				}
				stat = ZX_BAD_CONFIG_DATA;
				goto done;
			}
			recType = REC_TYPE_EEPROM;			
			buf[0] = cfgData[1];
			buf[1] = cfgData[2];
			addr = 0x0006;
			cnt = 3;
			dlen = 2;
			break;

		default:
			if (zxd.mesgFunc)
			{
				sprintf(msg, "unexpected configuration data type - %u\n", cfgData[0]);
				(*zxd.mesgFunc)(msg, zxd.mesgData);
			}
			stat = ZX_BAD_CONFIG_DATA;
			goto done;
			break;
		}

		// generate an Intel Hex format record for the configuration data
		zxd.bufLen += generateHexRecord(zxd.buf + zxd.bufLen, buf, dlen, addr, recType);
		zxd.recCnt++;
		cfgData += cnt;
		cfgLen -= cnt;
	}
	stat = (cfgLen == 0) ? ZX_SUCCESS : ZX_BAD_CONFIG_DATA;

	if (stat == ZX_SUCCESS)
	{
		// add the end record
		if (zxd.bufLen < bufLen)
		{
			zxd.bufLen += generateHexRecord(zxd.buf + zxd.bufLen, cfgData, 0, 0, REC_TYPE_END);
			zxd.recCnt++;
		}
		if (zxd.bufLen >= bufLen)
			stat = ZX_BAD_CONFIG_DATA;
	}

	if (stat == ZX_SUCCESS)
	{
		// get the device into the load mode
		if ((stat = prepareForDownload(&zxd, TRUE)) != ZX_SUCCESS)
			goto done;

		stat = sendHexRecords(&zxd, ZX_DOWNLOAD_WRITE | ZX_DOWNLOAD_CONFIG);

		// reset the device to cause it to begin executing
		if ((flags & ZX_NO_RESET) == 0)
			doReset(zxd.hand);
	}

done:
	// clean up
	if (zxd.ifp != NULL)
		fclose(zxd.ifp);
	if ((zxd.flags & CLOSE_PORT) && (zxd.hand != INVALID_HANDLE_VALUE))
		CloseHandle(zxd.hand);
	if (zxd.buf != NULL)
		free(zxd.buf);
	return(stat);
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXIdentify
 *
 * This exported entry point provides a means to request indentification
 * information from the ZX.  If the 'mesgFunc' callback is not null, it will
 * be called with message strings, typically to indicate an error condition.
 * The callback will also be invoked passing it a pointer to the identification
 * string.
 *
 * If the process completes successfully, the function will return ZX_SUCCESS.
 * Otherwise, a value will be returned to indicate the error condition (see
 * zxcomm.h for a list of values and their meanings).
 *
 */
int CALLBACK
ZXIdentify(long port, long flags, MesgFunc mesgFunc, long mesgData)
{
	int stat = ZX_SUCCESS;
	ZXCommData zxd;

	memset(&zxd, 0, sizeof(zxd));
	zxd.hand = INVALID_HANDLE_VALUE;
	zxd.mesgFunc = mesgFunc;
	zxd.mesgData = mesgData;

	// attempt to open the specified serial port
	if (flags & ZX_COMM_HANDLE)
		zxd.hand = (HANDLE)port;
	else if ((stat = commOpen(port, &zxd)) == ZX_SUCCESS)
		zxd.flags |= CLOSE_PORT;
	else
		goto done;

	// get the device into the load mode
	if ((stat = prepareForDownload(&zxd, TRUE)) != ZX_SUCCESS)
		goto done;

	// get the identification
	identifyDevice(&zxd);

	// reset the device to cause it to begin executing
	if ((flags & ZX_NO_RESET) == 0)
		doReset(zxd.hand);

done:
	// clean up
	if (zxd.ifp != NULL)
		fclose(zxd.ifp);
	if ((zxd.flags & CLOSE_PORT) && (zxd.hand != INVALID_HANDLE_VALUE))
		CloseHandle(zxd.hand);
	return(stat);
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXDownloadWnd
 *
 * This functions provides an alternate method of downloading code to a ZX but
 * using a window handle to provide a progress indication.  The lParam parameter
 * will be send to the window along with a value indicating the percentage
 * complete.
 *
 */
int CALLBACK
ZXDownloadWnd(const char *file, long port, long flags, HWND hWnd, UINT mesg, long lParam)
{
	CBData cbData;

	cbData.hWnd = hWnd;
	cbData.mesg = mesg;
	cbData.lParam = lParam;
	cbData.flags = flags;
	return(ZXDownload(file, port, flags, ZXHWndProgressCB, (long)&cbData));
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXUpdateWnd
 *
 * This functions provides an alternate method of updating firmware in a ZX but
 * using a window handle to provide a progress indication.  The lParam parameter
 * will be send to the window along with a value indicating the percentage
 * complete.
 *
 */
int CALLBACK
ZXUpdateWnd(const char *file, long port, long flags, HWND hWnd, UINT mesg, long lParam)
{
	CBData cbData;

	cbData.hWnd = hWnd;
	cbData.mesg = mesg;
	cbData.lParam = lParam;
	cbData.flags = flags;
	return(ZXUpdate(file, port, flags, ZXHWndProgressCB, (long)&cbData));
}

/*--------------------------------------------------------------------------*/

/*
 ** ZXHWndProgressCB
 *
 * This function is a callback that can be used to relay progress information
 * to a specified window via a message.  The wParam parameter of the message
 * will be the percentage complete expressed as a value between 0 and 100
 * inclusive.  The lParam of the message will be the value supplied by the
 * caller of the original function via its lParam parameter.
 *
 */
WORD CALLBACK
ZXHWndProgressCB(long recNum, long recCnt, long retryCnt, long data)
{
	CBData *cbData = (CBData *)data;
	if (cbData)
	{
		LRESULT result;
		WPARAM pctComplete;

		if (recCnt)
			pctComplete = (recNum * 100 + 50) / recCnt;
		else
			pctComplete = 0;
		result = SendMessage(cbData->hWnd, cbData->mesg, pctComplete, cbData->lParam);
		return((cbData->flags & ZX_IGNORE_HWND_RET) ? 1 : (result != 0));
	}
	return(1);
}

/*--------------------------------------------------------------------------*/

/** public functions **/

/*
 ** atn
 *
 * Toggle the DTR line as an attention signal.  The 'onTime' is
 * specified as a number of microseconds.
 *
 */
void
atn(HANDLE hand, WORD onTime)
{
	EscapeCommFunction(hand, SETDTR);
	if (onTime >= 10000)
		msDelay((WORD)(onTime / 1000));
	else
		usDelay(onTime);
	EscapeCommFunction(hand, CLRDTR);
}

/*
 ** clearQueue
 *
 * Remove all characters from the input queue.
 *
 */
void
clearQueue(HANDLE hand)
{
	while (1)
	{
		char c;
		DWORD cnt;
		COMSTAT cs;

		ClearCommError(hand, &cnt, &cs);
		if (cs.cbInQue == 0)
			break;
		ReadFile(hand, &c, 1, &cnt, NULL);
	}
}

/*
 ** commInit
 *
 * Open the specified COM port and prepare it for use.
 *
 */
int
commInit(unsigned port, HANDLE *handp)
{
	int stat = ZX_CANT_OPEN_PORT;
	char chan[20];
	HANDLE hand;

	// Form new style comm channel string - "\\.\COMn" where n is
	// the desired port number.  This allows for ports 10 and above.
	sprintf(chan, "\\\\.\\COM%u", port);

	// attempt to open the channel and set the state
	hand = CreateFile(chan, GENERIC_READ | GENERIC_WRITE, 0, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	if (hand != INVALID_HANDLE_VALUE)
	{
		DCB dcb;
		char commSpec[50];
		sprintf(commSpec, "COM%u:%u,N,8,1", port, STD_SPEED);

		memset(&dcb, 0, sizeof(dcb));
		dcb.DCBlength = sizeof(dcb);

		if (BuildCommDCB(commSpec, &dcb))
		{
			dcb.fBinary = 1;
			dcb.fDtrControl = dcb.fRtsControl = 0;

			if (SetCommState(hand, &dcb))
			{
				// ensure that DTR is not asserted
				EscapeCommFunction(hand, CLRDTR);
				stat = ZX_SUCCESS;
			}
		}
	}
	if ((stat != ZX_SUCCESS) && (hand != INVALID_HANDLE_VALUE))
	{
		CloseHandle(hand);
		hand = INVALID_HANDLE_VALUE;
	}
	if (handp != NULL)
		*handp = hand;
	return(stat ? ZX_CANT_OPEN_PORT : ZX_SUCCESS);
}

/*
 ** doAtn
 *
 * Attempt to gain control of the client system by toggling the DTR line.  If 'needResp'
 * is non-zero, await a a response from the device.  Otherwise, it is assumed that the
 * device is going to be ready.
 *
 * Return zero if successful, non-zero otherwise.
 *
 */
int
doAtn(HANDLE hand, int needResp)
{
	int i;

	// try signalling multiple times
	for (i = 0; i < ATN_RETRY; i++)
	{
		DWORD tick;
		int j;

		// pulse DTR high for 100us multiple times with 10ms spacing
		for (j = 0; j < ATN_COUNT; j++)
		{
			msDelay(10);
			atn(hand, RST_TIME);
		}

		if (!needResp)
			return(0);

		// await a response
		tick = GetTickCount() + ATN_TIMEOUT;
		while (1)
		{
			DWORD cnt;
			COMSTAT cs;
			BYTE c;

			// see if there is a character to process
			ClearCommError(hand, &cnt, &cs);
			if (cs.cbInQue == 0)
			{
				updateWindows(10);
				Yield();
			}
			else if (ReadFile(hand, &c, 1, &cnt, NULL) && (cnt > 0) && (c == PROMPT))
				return(0);

			// check for timeout expiration
			if (!(diagCode & DIAG_NO_TIME_LIMIT) && (GetTickCount() > tick))
				break;
		}
	}
	return(-1);
}

/*
 ** doReset
 *
 * Reset the device by toggling the DTR line.  Care must be
 * taken not to toggle the DTR line too many times or the
 * device will enter the load mode.
 *
 */
void
doReset(HANDLE cHand)
{
	int i;

	// try signalling multiple times
	for (i = 0; i < RST_COUNT; i++)
	{
		msDelay(10);
		atn(cHand, 100);
	}
}

/*
 ** hexVal
 *
 * Return the digit value of a hexadecimal character.  Zero is
 * returned for invalid hex characters.
 *
 */
BYTE
hexVal(BYTE c)
{
	if (isxdigit(c))
	{
		if (isdigit(c))
			return(c - '0');
		if (isupper(c))
			c = tolower(c);
		return(c - 'a' + 10);
	}
	return(0);
}

/*
 ** msDelay
 *
 * Delay for a specified number of milliseconds.
 *
 */
void
msDelay(WORD ms)
{
	if (ms > 65)
	{
		DWORD tick = GetTickCount() + ms;

		while (GetTickCount() < tick)
			;
	}
	else
		// for smaller values, to get better resolution
		usDelay((WORD)(ms * 1000));
}

/*
 ** setDiagCode
 *
 * This function, intended for use during debugging, sets a diagnostic
 * code that alters the behavior of certain parts of the code.
 *
 */
void
setDiagCode(int code)
{
	diagCode = code;
}

/*
 ** setSpeed
 *
 * Set the speed of the comm channel.  Return zero if successful.
 *
 */
int
setSpeed(HANDLE hand, DWORD speed)
{
	int stat = -1;
	DCB dcb;

	memset(&dcb, 0, sizeof(dcb));
	dcb.DCBlength = sizeof(dcb);

	if (GetCommState(hand, &dcb))
	{
		dcb.BaudRate = speed;
		if (SetCommState(hand, &dcb))
			stat = 0;
	}
	return(stat);
}

/*
 ** usDelay
 *
 * Delay for a specified number of microseconds.
 *
 */
void
usDelay(WORD us)
{
	__int64 end;

	if (QueryPerformanceCounter((LARGE_INTEGER *)&end))
	{
		__int64 freq, tick;

		// compute the number of counts to equal the requested delay
		QueryPerformanceFrequency((LARGE_INTEGER *)&freq);
		end += ((__int64)us * freq / 1000000);

		// implement the delay
		do
		{
			QueryPerformanceCounter((LARGE_INTEGER *)&tick);
		}
		while (tick <= end);
	}
}

/** private functions **/

/*
 ** commOpen
 *
 * Open the specified COM port and prepare it for use.
 *
 */
static int
commOpen(unsigned port, ZXCommData *zdp)
{
	int stat;

	if (zdp == NULL)
		return(ZX_PARAMETER_ERROR);

	stat = commInit(port, &zdp->hand);
	if (stat == ZX_CANT_OPEN_PORT)
	{
		char msg[100];
		sprintf(msg, "Can't open COM%u.", port);
		if (zdp->mesgFunc)
			(*zdp->mesgFunc)(msg, zdp->mesgData);
	}
	return(stat);
}

/*
 ** crc
 *
 * Compute a new CRC value given a current value and a new data byte.
 * The algorithm implemented is described by the RockSoft model parameters
 * shown below.  (see http://www.repairfaq.org/filipg/LINK/F_crc_v3.html)
 *
 * Name   :   "CRC-16"
 * Width  :   16
 * Poly   :   8005
 * Init   :   0000
 * RefIn  :   false
 * RefOut :   false
 * XorOut :   0000
 * Check  :   fee8
 *
 * Note, particularly, that the initial value of 'prevCRC' should be
 * zero when computing the CRC of a new stream of data.
 *
 */
static unsigned short
crc(unsigned char ch, unsigned short prevCRC)
{
#define POLYNOMIAL	0x8005L
	unsigned char i;

	unsigned long val24;
	
	// concatentate the input byte and the previous CRC value making a 24-bit value
	val24 = ((unsigned long)prevCRC << 8) | ch;

	for (i = 0; i < 8; i++)
	{
		// shift the 24-bit value left one place
		val24 <<= 1;

		// XOR the polynomial if bit 23 was a 1
		if (val24 & 0x1000000L)
			val24 ^= (POLYNOMIAL << 8);
	}

	// return bits 8-23 of the result as the new CRC value
	return((unsigned short)(val24 >> 8));
}

/*
 ** generateHexRecord
 *
 * Generate an intel Hex record for the data provided.  Return the
 * number of characters added to the record buffer.
 *
 */
static int
generateHexRecord(char *recBuf, const char *dataBuf, unsigned cnt, unsigned addr, char recType)
{
	int len;
	char cksum;
	unsigned i;

	if ((recBuf == NULL) || (dataBuf == NULL))
		return(0);

	addr &= 0xffff;
	len = sprintf(recBuf, ":%02x%04x%02x", cnt & 0xff, addr, recType);
	cksum = cnt + addr + (addr >> 8) + recType;
	for (i = 0; i < cnt; i++)
	{
		len += sprintf(recBuf + len, "%02x", dataBuf[i]);
		cksum += dataBuf[i];
	}
	len += sprintf(recBuf + len, "%02x\n", -(int)cksum & 0xff);
	return(len);
}

/*
 ** getRecordByte
 *
 * Extract a byte from a buffer that is expected to contain a valid
 * Intel hex format record.
 *
 */
static BYTE
getRecordByte(const BYTE *buf, int idx)
{
	BYTE hiByte, loByte;

	if (!isxdigit(hiByte = buf[idx]) || !isxdigit(loByte = buf[idx + 1]))
		return(0);
	return((hexVal(hiByte) << 4) | hexVal(loByte));
}

/*
 ** getResponse
 *
 * Await a response from the device after a record has been sent.  Return
 * zero if the response is acceptable, non-zero otherwise.
 *
 */
static int
getResponse(ZXCommData *zdp, long recCnt)
{
	DWORD tick;

	if ((zdp == NULL) || (zdp->hand == INVALID_HANDLE_VALUE))
		return(0);

	// wait for the acknowledgement
	tick = GetTickCount() + 1500;
	while (1)
	{
		char msg[100];
		DWORD cnt;
		COMSTAT cs;
		BYTE c;

		if (!(diagCode & DIAG_NO_TIME_LIMIT) && (GetTickCount() > tick))
		{
			if (zdp->mesgFunc)
			{
				sprintf(msg, "Device failed to timely acknowledge record %lu.\n", recCnt);
				(*zdp->mesgFunc)(msg, zdp->mesgData);
			}
			return(1);
		}

		ClearCommError(zdp->hand, &cnt, &cs);
		if (cs.cbInQue == 0)
		{
			updateWindows(10);
			Yield();
		}
		else if (ReadFile(zdp->hand, &c, 1, &cnt, NULL) && (cnt > 0))
		{
			switch(c)
			{
			case EOF_CHAR:
			case ACK_CHAR:
			case RST_CHAR:
				return(0);
				break;

			case CRC_CHAR:
				if (zdp->mesgFunc)
				{
					sprintf(msg, "Device reports CRC error, record %lu.", recCnt);
					(*zdp->mesgFunc)(msg, zdp->mesgData);
				}
				return(1);
				break;

			case SIG_CHAR:
				if (zdp->mesgFunc)
				{
					sprintf(msg, "Device reports bad signature, record %lu.", recCnt);
					(*zdp->mesgFunc)(msg, zdp->mesgData);
				}
				return(1);
				break;

			case TIMEOUT_CHAR:
				return(2);
				break;

			default:
				if (zdp->mesgFunc)
				{
					sprintf(msg, "Unexpected device response - 0x%02x, record %lu.", (int)c & 0xff, recCnt);
					(*zdp->mesgFunc)(msg, zdp->mesgData);
				}
				return(1);
				break;
			}
		}
	}
	return(1);
}

/*
 ** identifyDevice
 *
 * Retrieve and display an identification string from the device.  It is assumed
 * that the device is already in command mode.
 *
 */
static int
identifyDevice(ZXCommData *zdp)
{
	char buf[200], c, *p, *p1;
	DWORD cnt;
	int idx = 0;

	if ((zdp == NULL) || (zdp->hand == INVALID_HANDLE_VALUE))
		return(ZX_PARAMETER_ERROR);

	// send the identification command to the device, collect the response
	WriteFile(zdp->hand, "I\r", 2, &cnt, NULL);
	msDelay(150);
	while (ReadFile(zdp->hand, &c, 1, &cnt, NULL) && (cnt > 0) && (idx < (sizeof(buf) - 1)))
		buf[idx++] = c;
	buf[idx] = '\0';

	// skip over leading whitespace, command echo, etc.
	p = skipWhite(buf);
	if (*p == 'I')
		p = skipWhite(p + 1);
	if (((p1 = strchr(p, '\r')) != NULL) && (p1[1] == '\n'))
	{
		if (p1[2] != PROMPT)
		{
			*p1++ = ' ';
			strcpy(p1, p1 + 1);
			p1 = strchr(p1, PROMPT);
		}
		if (p1 != NULL)
			*p1 = '\0';
	}

	if (zdp->mesgFunc)
		(*zdp->mesgFunc)(p, zdp->mesgData);
	return(ZX_SUCCESS);
}

/*
 ** prepareForDownload
 *
 * Prepare the device for downloading using the given comm handle.
 * If successful, return zero.  Otherwise, return an error code.
 *
 *
 */
static int
prepareForDownload(ZXCommData *zdp, int needAtnResponse)
{
	HANDLE hand;

	if ((zdp == NULL) || ((hand = zdp->hand) == INVALID_HANDLE_VALUE))
		return(ZX_PARAMETER_ERROR);

	// deassert DTR
	EscapeCommFunction(hand, CLRDTR);

	if ((DL_SPEED != STD_SPEED) && (setSpeed(hand, DL_SPEED) != 0))
	{
		DWORD err = GetLastError();
		char msg[100];
		if (zdp->mesgFunc)
		{
			sprintf(msg, "Attempt to change to download speed failed, err=%lu.", err);
			(*zdp->mesgFunc)(msg, zdp->mesgData);
		}
		return(ZX_FAILED_SPEED_CHANGE);
	}

	// get the device into the load mode
	if (doAtn(hand, needAtnResponse) != 0)
	{
		if (zdp->mesgFunc)
			(*zdp->mesgFunc)("Device failed to respond to the ATN signal.", zdp->mesgData);
		return(ZX_NO_ATN_REPSONSE);
	}

	return(ZX_SUCCESS);
}

/*
 ** sendHexData
 *
 * Send a sequence of Intel hex records out the serial channel, record by record.
 * Return zero if successful, non-zero otherwise.
 *
 */
static int
sendHexRecords(ZXCommData *zdp, unsigned flags)
{
#define TX_RETRIES		3
	DWORD cnt;
	DWORD tick;
	COMSTAT cs;
	BYTE c;
	unsigned bufIdx = 0;
	char buf[100];
	char msg[150];
	long recCnt = 0;
	int dotCnt = 0;
	char *op;
	unsigned mode = flags & ZX_MODE_MASK;
	HANDLE hand;

	if ((zdp == NULL) || ((hand = zdp->hand) == INVALID_HANDLE_VALUE))
		return(ZX_PARAMETER_ERROR);

	if (zdp->ifp != NULL)
	{
		// rewind the file
		fseek(zdp->ifp, 0, 0);
	}
	else if ((zdp->buf == NULL) || !zdp->bufLen)
		// no data to send
		return(ZX_PARAMETER_ERROR);

	// send the load or verify command depending on the flags
	sprintf(buf, "%c\r", (mode == ZX_DOWNLOAD_VERIFY) ? 'V' : 'L');
	WriteFile(hand, buf, strlen(buf), &cnt, NULL);

	// wait for the ACK
	tick = GetTickCount() + 1000;

	op = (mode == ZX_DOWNLOAD_VERIFY) ? "verify" : ((flags & ZX_DOWNLOAD_CONFIG) ? "configuration" : "download");
	while (1)
	{
		ClearCommError(hand, &cnt, &cs);
		if (!(diagCode & DIAG_NO_TIME_LIMIT) && (GetTickCount() > tick))
		{
			if (zdp->mesgFunc)
			{
				sprintf(msg, "No %s ack from device.", op);
				(*zdp->mesgFunc)(msg, zdp->mesgData);
			}
			return(ZX_COMM_TIMEOUT);
		}
		if (cs.cbInQue == 0)
		{
			updateWindows(10);
			Yield();
		}
		else if (ReadFile(hand, &c, 1, &cnt, NULL) && (cnt > 0) && (c == ACK_CHAR))
			break;
	}
	if ((diagCode & DIAG_VERBOSE) && zdp->mesgFunc)
	{
		sprintf(msg, "Received ack for %s command.\n", op);
		(*zdp->mesgFunc)(msg, zdp->mesgData);
	}

	while (1)
	{
		int len, txCnt;
		int waitAck;
		unsigned char recType;

		if (zdp->ifp != NULL)
		{
			// read the next record from the file
			if (fgets(buf, sizeof(buf), zdp->ifp) == NULL)
				*buf = '\0';
		}
		else
		{
			// copy the next record to the buffer
			char *p = buf;
			while (bufIdx < zdp->bufLen)
			{
				if ((*p++ = zdp->buf[bufIdx++]) == '\n')
					break;
			}
			*p = '\0';
		}
		if (*buf == '\0')
			break;

		// remove whitespace from the end
		len = trimWhite(buf);
		recType = ((len >= 9) ? getRecordByte(buf, 7) : 0);

		// set for multiple transmission attempts
		txCnt = TX_RETRIES;
sendRecord:

		// purge the incoming queue
		clearQueue(hand);

		if (txCnt == TX_RETRIES)
		{
			recCnt++;
			if (zdp->progFunc && !(*zdp->progFunc)(recCnt, zdp->recCnt, 0, zdp->progData))
				return(ZX_ABORTED);
		}

		// send the record
		WriteFile(hand, buf, len, &cnt, NULL);

		// wait for acknowledgement
		waitAck = TRUE;
		tick = GetTickCount() + 1000;
		while (waitAck)
		{
			if (!(diagCode & DIAG_NO_TIME_LIMIT) && (GetTickCount() > tick))
			{
				int stat;
				if ((mode != ZX_DOWNLOAD_VERIFY) && (recType == REC_TYPE_FIRMWARE))
				{
					strcpy(msg, "No response to firmware version check.  Please update your firmware.");
					stat = ZX_VERSION_TIMEOUT;
				}
				else
				{
					sprintf(msg, "No response from device, record %u.", recCnt);
					stat = ZX_COMM_TIMEOUT;
				}
				if (zdp->mesgFunc)
					(*zdp->mesgFunc)(msg, zdp->mesgData);
				return(stat);
			}
			ClearCommError(hand, &cnt, &cs);
			if (cs.cbInQue == 0)
			{
				updateWindows(10);
				Yield();
			}
			else if (ReadFile(hand, &c, 1, &cnt, NULL) && (cnt > 0))
			{
				switch(c)
				{
				case ' ':
				case '\t':
				case '\r':
				case '\n':
					continue;

				case VER_CHAR:
					if ((mode != ZX_DOWNLOAD_VERIFY) && (recType == REC_TYPE_FIRMWARE))
					{
						if (zdp->mesgFunc)
						{
							sprintf(msg, "Device failed firmware version check, v%u.%u.%u or later is required.",
									getRecordByte(buf, 9), getRecordByte(buf, 11), getRecordByte(buf, 13));
							(*zdp->mesgFunc)(msg, zdp->mesgData);
						}
						return(ZX_VERSION_FAILURE);
					}
					// continue into next case

				case NAK_CHAR:
					if (--txCnt > 0)
						goto sendRecord;
					if (zdp->mesgFunc)
					{
						sprintf(msg, "%s, record %u.", (c == NAK_CHAR) ?
							"Negative acknowledgement from device" : "Verification error", recCnt - 1);
						(*zdp->mesgFunc)(msg, zdp->mesgData);
					}
					return(ZX_RECORD_NAK);
					break;

				case ACK_CHAR:
					if ((diagCode & DIAG_VERBOSE) && zdp->mesgFunc)
					{
						sprintf(msg, "Affirmative response, record %u.", recCnt);
						(*zdp->mesgFunc)(msg, zdp->mesgData);
					}
					waitAck = FALSE;
					break;

				default:
					// unexpected response
					if (zdp->mesgFunc)
					{
						sprintf(msg, "Unexpected device response (0x%02x), record %u.\r\n", c, recCnt);
						(*zdp->mesgFunc)(msg, zdp->mesgData);
					}
					break;
				}
			}
		}
	}

	if (zdp->mesgFunc)
	{
		sprintf(msg, "%s completed successfully.",
				(mode == ZX_DOWNLOAD_VERIFY) ? "Verification" : ((flags & ZX_DOWNLOAD_CONFIG) ? "Configuration" : "Download"));
		(*zdp->mesgFunc)(msg, zdp->mesgData);
	}
	return(ZX_SUCCESS);
}

/*
 ** skipWhite
 *
 * Advance a pointer to the first non-whitespace character.
 *
 */
static char *
skipWhite(const char *p)
{
	if (p != NULL)
	{		
		char c;
		while (((c = *p) != '\0') && isspace(c))
			p++;
	}
	return((char *)p);
}

/*
 ** trimWhite
 *
 * Remove whitespace from the end of a string.
 *
 */
static int
trimWhite(char *str)
{
	int len;
	for (len = strlen(str); len > 0; len--)
		if (!isspace(str[len - 1]))
			break;
	str[len] = '\0';
	return(len);
}

/*
 ** updateFirmware
 *
 * Perform a firmware update.  Return zero if successful, non-zero otherwise.
 * See verifyUpdate() for a description of the file content.
 *
 * It is assumed that the integrity of the file has already been checked.
 *
 */
static int
updateFirmware(ZXCommData *zdp, unsigned flags, int *retryp)
{
#define MAX_REC_RETRIES		5
#define UPDATE_COMPLETE		1000
	int stat = 0;
	char buf[200];
	char recBuf[512];
	unsigned char c;
	unsigned byteCnt = 0;
	unsigned idx = 0;
	unsigned recIdx = 0;
	int state = 0;
	DWORD cnt;
	COMSTAT cs;
	long recCnt = 0;
	int dotCnt = 0;
	int recRetryCnt = 0;
	int retryCnt = 0;
	int lastRetryCnt = 0;
	long recStart = 0;
	long bufStart = 0;
	HANDLE hand;

	if ((zdp == NULL) || ((hand = zdp->hand) == INVALID_HANDLE_VALUE) || (zdp->ifp == NULL))
		return(ZX_PARAMETER_ERROR);

	// rewind the file
	fseek(zdp->ifp, 0, 0);

	if (!(flags & ZX_UPDATE_EMERGENCY))
	{
		// send the update command
		sprintf(buf, "%c\r", (flags & ZX_UPDATE_NO_BLINK) ? 'u' : 'U');
		WriteFile(hand, buf, strlen(buf), &cnt, NULL);
	}

	// delay a bit to allow the loader to initialize
	msDelay(250);

	// empty the input queue
	clearQueue(hand);

	// send the update data
	while (stat == 0)
	{
		unsigned recLen;
		unsigned short val;

		// read another line if necessary
		if (idx >= byteCnt)
		{
			// save the offset of the beginning of buffer
			bufStart = ftell(zdp->ifp);

			// read another line
			if (fgets(buf, sizeof(buf), zdp->ifp) == NULL)
				break;
			idx = 0;

			// remove whitespace from the end of the line
			byteCnt = trimWhite(buf);
			if (byteCnt == 0)
				continue;

			// ignore lines not beginning with a hex digit
			if (!isxdigit(*buf))
			{
				byteCnt = 0;
				continue;
			}
		}

		// get the next character from the buffer, convert to a value
		c = hexVal(buf[idx++]);
		switch(state)
		{
		case 0:		// reading a record length
			// save the offset of the beginning of the  record and initialize the record length value
			recStart  = bufStart + idx - 1;
			recLen = 0;
			// continue into next state

		case 1:
		case 2:
		case 3:
			recLen = (recLen << 4) + c;
			if (state++ >= 3)
			{
				if (recLen == 0)
					stat = UPDATE_COMPLETE;
				else
				{
					// send the record length, MSB first
					recBuf[0] = recLen >> 8;
					recBuf[1] = recLen;
					WriteFile(hand, recBuf, 2, &cnt, NULL);
					recIdx = 0;
				}
			}
			break;

		case 4:		// reading a record byte
		case 5:
			if (state == 4)
			{
				val = c;
				state = 5;
			}
			else
			{
				// put the byte in the record buffer
				recBuf[recIdx++] = (val << 4) | c;
				if (recIdx < recLen)
				{
					state = 4;
					break;
				}

				// see if a premature response has been received
				ClearCommError(hand, &cnt, &cs);
				if (cs.cbInQue == 0)
				{
					// send the record's bytes to the device
					WriteFile(hand, recBuf, recLen, &cnt, NULL);

					stat = getResponse(zdp, recCnt);
					if (stat == 0)
					{
						// record successfully sent
						if (recRetryCnt)
							retryCnt++;
						recRetryCnt = 0;

						// update the progress indicator
						if (zdp->progFunc)
							(*zdp->progFunc)(++recCnt, zdp->recCnt, retryCnt, zdp->progData);
						state = 0;
						break;
					}
				}

				// an error has occurred or an unexpected response was received, attempt a retry
				if (++recRetryCnt <= MAX_REC_RETRIES)
				{
					// prepare to re-transmit the record
					fseek(zdp->ifp, recStart, SEEK_SET);
					state = 0;
					idx = byteCnt = 0;
		 			msDelay(100);
					clearQueue(hand);
					stat = 0;
				}
				else
				{
					if (zdp->mesgFunc)
						(*zdp->mesgFunc)("Retry limit exceeded, aborting update.", zdp->mesgData);
					stat = ZX_RETRY_LIMIT_EXCEEDED;
				}
			}
			break;

		default:
			// this shouldn't happen
			stat = 1;
			break;
		}
	}

	// check for a successful update
	if (stat == UPDATE_COMPLETE)
	{
		if (zdp->progFunc)
			(*zdp->progFunc)(zdp->recCnt, zdp->recCnt, 0, zdp->progData);
		stat = 0;
	}
	if (retryp != NULL)
		*retryp = retryCnt;
	return(stat);
}

/*
 ** updateWindows
 *
 * Process Windows messages.
 *
 */
static void
updateWindows(unsigned maxMsgs)
{
#ifdef _WINDOWS
	MSG msg;
	unsigned msgCnt = 0;
	while (PeekMessage(&msg, NULL, 0, UINT_MAX, PM_NOREMOVE))
	{
		if (GetMessage(&msg, NULL, 0, 0))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		if (maxMsgs && (++msgCnt >= maxMsgs))
			break;
	}
#endif
}

/*
 ** validateHexRecord
 *
 * Scan an Intel hex record and validate its format.  Return
 * the record type if it is valid and return the length (in characters)
 * indirectly.  Otherwise, return an invalid record type.
 *
 */
static BYTE
validateHexRecord(const BYTE *buf, int *lenp)
{
	int state;
	BYTE c;
	BYTE byteCnt = 0;
	BYTE byteVal = 0;
	BYTE charCnt = 0;
	BYTE cksum = 0;
	BYTE recType = 0;
	WORD addr = 0;

	state = HEX_READ_START;
	while (state != HEX_READ_DONE)
	{
		c = *buf++;
		if (state == HEX_READ_START)
		{
			if (c != ':')
				return(REC_TYPE_INVALID);

			state = HEX_READ_BYTECNT;
			cksum = 0;
			byteCnt = 0;
			charCnt = 2;
			continue;
		}

		// otherwise we need a hex digit
		if (!isxdigit(c))
			return(REC_TYPE_INVALID);
		c = hexVal(c);
		switch(state)
		{
		case HEX_READ_BYTECNT:		// reading byte count
			byteCnt = (byteCnt << 4) + c;
			if (--charCnt <= 0)
			{
				cksum += byteCnt;
				state = HEX_READ_ADDRESS;
				addr = 0;
				charCnt = 4;
			}
			break;

		case HEX_READ_ADDRESS:	// reading address
			addr = (addr << 4) + c;
			if (--charCnt <= 0)
			{
				cksum += (((addr >> 8) & 0xff) + (addr & 0xff));
				state = HEX_READ_RECTYPE;
				recType = 0;
				charCnt = 2;
			}
			break;

		case HEX_READ_RECTYPE:	// reading record type
			recType = (recType << 4) + c;
			if (--charCnt <= 0)
			{
				cksum += recType;
				byteVal = 0;
				if (byteCnt > 0)
				{
					state = HEX_READ_RECORD;
					charCnt = byteCnt * 2;
				}
				else
				{
					state = HEX_READ_CHKSUM;
					charCnt = 2;
				}
			}
			break;

		case HEX_READ_RECORD:	// reading record bytes
			byteVal = (byteVal << 4) + c;
			if ((--charCnt & 0x01) == 0)
			{
				cksum += byteVal;
				byteVal = 0;
			}
			if (charCnt <= 0)
			{
				state = HEX_READ_CHKSUM;
				byteVal = 0;
				charCnt = 2;
			}
			break;

		case HEX_READ_CHKSUM:	// reading checksum
			byteVal = (byteVal << 4) + c;
			if (--charCnt <= 0)
			{
				if (((cksum + byteVal) & 0xff) != 0)
					return(REC_TYPE_INVALID);
				state = HEX_READ_DONE;
			}
			break;

		default:	// this shouldn't ever happen
			return(REC_TYPE_INVALID);
		}
	}

	// confirm that no non-whitespace characters follow the checksum
	while (1)
	{
		if ((c = *buf++) == '\0')
			break;
		if (!isspace(c))
			return(REC_TYPE_INVALID);
	}

	if (lenp != NULL)
		*lenp = (byteCnt * 2) + 11;
	return(recType);
}

/*
 ** verifyIntelHex
 *
 * Verify that a file or data stream contains a valid Intel hex download image.
 * While doing so, count the number of records in the image.
 *
 */
static int
verifyIntelHex(ZXCommData *zdp)
{
	unsigned bufIdx = 0;
	char buf[100];

	if ((zdp == NULL) || ((zdp->ifp == NULL) && (zdp->buf == NULL)))
		return(FALSE);

	zdp->recCnt = 0;
	for(;;)
	{
		// get the next record to process
		if (zdp->ifp != NULL)
		{
			// ... from the file
			if (fgets(buf, sizeof(buf), zdp->ifp) == NULL)
				*buf = '\0';
		}
		else
		{
			// ... from the buffer
			char *p = buf;
			while (bufIdx < zdp->bufLen)
			{
				if ((*p++ = zdp->buf[bufIdx++]) == '\n')
					break;
			}
			*p = '\0';
		}
		if (*buf == '\0')
			break;

		// see if the record is properly formatted
		if (validateHexRecord(buf, NULL) == REC_TYPE_INVALID)
			return(FALSE);
		zdp->recCnt++;
#ifdef _WINDOWS
		if ((zdp->recCnt % 10) == 0)
			updateWindows(0);
#endif
	}
	if (zdp->ifp != NULL)
		fseek(zdp->ifp, 0, 0);
	return(TRUE);
}

/*
 ** verifyUpdate
 *
 * Verify the integrity of an update file.  Return non-zero if
 * it's OK, zero otherwise.
 *
 * The update file is expected to have records consisting
 * of a record length (4 hex characters) followed by the
 * bytes of the record (2 hex characters each).  The records
 * may be split across lines.  The final record of the file
 * should have a zero length and be followed by the CRC
 * value of all of the preceeding record data (not including
 * the zero record length).  The CRC is four hex characters.
 *
 * Any line not beginning with a hex character is assumed
 * to be a comment and is ignored.  Blank lines are ignored.
 *
 */
static int
verifyUpdate(ZXCommData *zdp)
{
	int stat = TRUE;
	unsigned short crcVal = 0;
	unsigned short crcLast;
	char buf[200];
	unsigned recLen = 0;
	unsigned byteCnt = 0;
	unsigned idx = 0;
	int state = 0;

	if ((zdp == NULL) || ((zdp->ifp == NULL) && (zdp->buf == NULL)))
		return(FALSE);
	zdp->recCnt = 0;

	// do a quick scan of the file to check its integrity
	while (stat)
	{
		unsigned char c;
		unsigned short val;

		// read another line if necessary
		if (idx >= byteCnt)
		{
			// check for a byte being split
			if (state & 1)
			{
				stat = 0;
				break;
			}

			// read another line
			if (fgets(buf, sizeof(buf), zdp->ifp) == NULL)
				break;
			idx = 0;

			// remove whitespace from the end of the line
			byteCnt = trimWhite(buf);
			if (byteCnt == 0)
				continue;

			// ignore lines not beginning with a hex digit
			if (!isxdigit(*buf))
			{
				byteCnt = 0;
				continue;
			}
		}

		c = buf[idx++];
		if (!isxdigit(c))
		{
			// invalid character
			stat = 0;
			break;
		}
		if (state == 0)
			crcLast = crcVal;
		crcVal = crc(c, crcVal);

		// convert to a value
		c = hexVal(c);
		switch(state)
		{
		case 0:		// reading a record length
		case 1:
		case 2:
		case 3:
			if (state == 0)
				recLen = c;
			else
				recLen = (recLen << 4) + c;
			if ((state == 3) && (recLen == 0))
				// zero record length indicates that the CRC follows
				state = 6;
			else
				state++;
			break;

		case 4:		// reading a record byte
		case 5:
			if (state == 4)
				state = 5;
			else
			{
				// a complete byte has been read
				if (--recLen == 0)
				{
					// the record has been read
					state = 0;
					zdp->recCnt++;
#ifdef _WINDOWS
					if ((zdp->recCnt % 10) == 0)
						updateWindows(0);
#endif
				}
				else
					state = 4;
			}
			break;

		case 6:		// reading the CRC value
		case 7:
		case 8:
		case 9:
			if (state == 6)
				val = c;
			else
				val = (val << 4) + c;
			if (state == 9)
			{
				// check the computed CRC against that in the file
				if (val != crcLast)
					stat = 0;
				state = 10;
			}
			else
				state++;
			break;

		default:
			// this shouldn't happen
			stat = 0;
			break;
		}
	}
	if (zdp->ifp != NULL)
		fseek(zdp->ifp, 0, 0);
	return(stat);
}


